/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid Touati

   This file is part of GDD.
   
   GDD is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   GDD is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with GDD.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <libxml/parser.h>

#include <stdarg.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>
#include <GDD/ddg.h>

#include <assert.h>

void usage(const char *app) {
  fprintf(stdout,"%s -a arch.xml -i ddg.xml\n", app);
}

#define WARNING 0
#define ERROR 1

void warn_error(int status, const char *app, const char *fmt, ...) {
  va_list ap;
  if(status == WARNING) {
    fprintf(stderr,"warning: ");
  } else {
    fprintf(stderr,"error: ");
  }
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fprintf(stderr,"\n");
  usage(app);
  if(status != WARNING) {
    exit(status);
  }
}

double ellapsed_time(struct timeval *t0, struct timeval *t1) {
  return ((double)(t1->tv_sec - t0->tv_sec)) + ((double)(t1->tv_usec - t0->tv_usec))/((double)1000000);
}

int main(int argc, char *argv[]) {
  GDD_Arch *arch = NULL;
  SCEDA_Graph *g = NULL;
  int IImin_rec;

  xmlInitParser();

  int i;
  for(i = 0; i < argc; i++) {
    if(strcmp(argv[i],"-i") == 0) {
      i++;
      if(i < argc) {
	if(g == NULL) {
	  char *filename = argv[i];
	  if(arch == NULL) {
	    fprintf(stderr,"warning: no arch file given\n");
	    fprintf(stderr,"attempting to load default arch file\n");
	    arch = GDD_arch_of_xml_file("default_arch.xml");
	    assert(arch != NULL);
	  }
	  fprintf(stderr,"loading ddg file %s\n",filename);
	  g = GDD_of_xml_file(FALSE, arch, filename, &IImin_rec);
	  assert(g != NULL);
	} else {
	  warn_error( ERROR, argv[0], "ddg file already given");
	}
      } 
    } else if(strcmp(argv[i],"-a") == 0) {
      i++;
      if(i < argc) {
	if(arch == NULL) {
	  char *filename = argv[i];
	  fprintf(stderr,"loading arch file %s\n",filename);
	  arch = GDD_arch_of_xml_file(filename);
	  assert(arch != NULL);
	} else {
	  warn_error( ERROR, argv[0], "arch file already given");
	}
      }
    } 
  }

  xmlCleanupParser();

  if(g == NULL) {
    warn_error( ERROR, argv[0], "no ddg file given");
  }

  struct timeval t0;
  struct timeval t1;

  SCEDA_List *cycle;

  gettimeofday(&t0, NULL);
  int is_lex_pos = GDD_is_lexicographic_positive(g, &cycle);
  gettimeofday(&t1, NULL);

  double bench1 = ellapsed_time(&t0, &t1);

  if(!is_lex_pos) {
    SCEDA_list_delete(cycle);
    fprintf(stdout,"error: the given GDD is not lexicographic positive\n");
  }
  
  int MIInum, MIIden;

  gettimeofday(&t0, NULL);
  int has_MII = GDD_MII_ratio_cycle(g, &MIInum, &MIIden, &cycle);
  gettimeofday(&t1, NULL);

  double bench2 = ellapsed_time(&t0, &t1);

  fprintf(stdout,"ellapsed time=\t%f\t%f\n",bench1, bench2);

  if(has_MII == 0) {
    fprintf(stdout,"MII=%d/%d\n", MIInum, MIIden);

    fprintf(stdout,"Here is a cycle that reaches the MII:\n");
    {
      // print the cycle
      SCEDA_ListIterator edges;
      SCEDA_list_iterator_init(cycle, &edges);
      while(SCEDA_list_iterator_has_next(&edges)) {
	SCEDA_Edge *e = SCEDA_list_iterator_next(&edges);
	SCEDA_Vertex *vs = SCEDA_edge_source(e);
	SCEDA_Vertex *vt = SCEDA_edge_target(e);
	fprintf(stderr,"%s -(%d,%d)-> %s\n", GDD_vertex_mnemonic(vs), GDD_edge_latency(e), GDD_edge_distance(e), GDD_vertex_mnemonic(vt));
      }
      SCEDA_list_iterator_cleanup(&edges);
    }

    SCEDA_ListIterator MIIcycle;
    SCEDA_list_iterator_init(cycle, &MIIcycle);
    assert(SCEDA_list_iterator_has_next(&MIIcycle));
    SCEDA_Edge *e = SCEDA_list_iterator_next(&MIIcycle);
    SCEDA_Vertex *source = SCEDA_edge_source(e);
    SCEDA_Vertex *target = SCEDA_edge_target(e);
    int distance = GDD_edge_distance(e);

    SCEDA_Edge *en = SCEDA_graph_add_edge(g, source, target, GDD_edge_label_create(NULL, "serial", 0, distance - MIIden));

    SCEDA_list_iterator_cleanup(&MIIcycle);

    SCEDA_list_delete(cycle);

    if(GDD_is_lexicographic_positive(g, &cycle)) {
      fprintf(stdout,"error: just discovered a bug! Please report!\n");
    } else {
      SCEDA_list_delete(cycle);
    }

    GDD_EdgeLabel *data;
    SCEDA_graph_remove_edge(g, en, (void **)&data);
    GDD_edge_label_delete(data);
  }

  SCEDA_graph_delete(g);
  GDD_arch_delete(arch);
  return 0;
}
